"""Integrate Betty with `nginx <https://nginx.org/>`_."""
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout, QButtonGroup, QRadioButton, QWidget, QHBoxLayout, QLineEdit, \
    QFileDialog, QPushButton

from betty.app import App
from betty.app.extension import ConfigurableExtension
from betty.config import Configuration
from betty.extension.nginx.artifact import generate_configuration_file, generate_dockerfile_file
from betty.generate import Generator, GenerationContext
from betty.gui import GuiBuilder
from betty.gui.error import ExceptionCatcher
from betty.locale import Str
from betty.serde.dump import Dump, VoidableDump, minimize, Void, VoidableDictDump
from betty.serde.load import Asserter, Fields, OptionalField, Assertions
from betty.serve import ServerProvider, Server


class NginxConfiguration(Configuration):
    def __init__(
        self,
        *,
        www_directory_path: str | None = None,
        https: bool | None = None,
    ):
        super().__init__()
        self._https = https
        self.www_directory_path = www_directory_path

    @property
    def https(self) -> bool | None:
        return self._https

    @https.setter
    def https(self, https: bool | None) -> None:
        self._https = https
        self._dispatch_change()

    @property
    def www_directory_path(self) -> str | None:
        return self._www_directory_path

    @www_directory_path.setter
    def www_directory_path(self, www_directory_path: str | None) -> None:
        self._www_directory_path = www_directory_path
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._https = other._https
        self._www_directory_path = other._www_directory_path
        self._dispatch_change()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'https',
                Assertions(asserter.assert_or(asserter.assert_bool(), asserter.assert_none())) | asserter.assert_setattr(configuration, 'https'),
            ),
            OptionalField(
                'www_directory_path',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'www_directory_path'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        dump: VoidableDictDump[VoidableDump] = {
            'https': self.https,
            'www_directory_path': Void if self.www_directory_path is None else str(self.www_directory_path),
        }
        return minimize(dump, True)


class _Nginx(ConfigurableExtension[NginxConfiguration], Generator, ServerProvider, GuiBuilder):
    @classmethod
    def label(cls) -> Str:
        return Str.plain('Nginx')

    @classmethod
    def description(cls) -> Str:
        return Str._('Generate <a href="">nginx</a> configuration for your site, as well as a <code>Dockerfile</code> to build a <a href="https://www.docker.com/">Docker</a> container around it.')

    @classmethod
    def default_configuration(cls) -> NginxConfiguration:
        return NginxConfiguration()

    @property
    def servers(self) -> Sequence[Server]:
        from betty.extension.nginx.serve import DockerizedNginxServer

        if DockerizedNginxServer.is_available():
            return [DockerizedNginxServer(self._app)]
        return []

    async def generate(self, job_context: GenerationContext) -> None:
        await generate_configuration_file(self._app)
        await generate_dockerfile_file(self._app)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @property
    def https(self) -> bool:
        if self._configuration.https is None:
            return self._app.project.configuration.base_url.startswith('https')
        return self._configuration.https

    @property
    def www_directory_path(self) -> str:
        return self._configuration.www_directory_path or str(self._app.project.configuration.www_directory_path)

    def gui_build(self) -> QWidget:
        return _NginxGuiWidget(self._app, self._configuration)


class _NginxGuiWidget(QWidget):
    def __init__(self, app: App, configuration: NginxConfiguration, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._app = app
        self._configuration = configuration
        layout = QFormLayout()

        self.setLayout(layout)

        https_button_group = QButtonGroup()

        def _update_configuration_https_base_url(checked: bool) -> None:
            if checked:
                self._configuration.https = None
        self._nginx_https_base_url = QRadioButton("Use HTTPS and HTTP/2 if the site's URL starts with https://")
        self._nginx_https_base_url.setChecked(self._configuration.https is None)
        self._nginx_https_base_url.toggled.connect(_update_configuration_https_base_url)
        layout.addRow(self._nginx_https_base_url)
        https_button_group.addButton(self._nginx_https_base_url)

        def _update_configuration_https_https(checked: bool) -> None:
            if checked:
                self._configuration.https = True
        self._nginx_https_https = QRadioButton('Use HTTPS and HTTP/2')
        self._nginx_https_https.setChecked(self._configuration.https is True)
        self._nginx_https_https.toggled.connect(_update_configuration_https_https)
        layout.addRow(self._nginx_https_https)
        https_button_group.addButton(self._nginx_https_https)

        def _update_configuration_https_http(checked: bool) -> None:
            if checked:
                self._configuration.https = False
        self._nginx_https_http = QRadioButton('Use HTTP')
        self._nginx_https_http.setChecked(self._configuration.https is False)
        self._nginx_https_http.toggled.connect(_update_configuration_https_http)
        layout.addRow(self._nginx_https_http)
        https_button_group.addButton(self._nginx_https_http)

        def _update_configuration_www_directory_path(www_directory_path: str) -> None:
            self._configuration.www_directory_path = None if www_directory_path == '' or www_directory_path == str(self._app.project.configuration.www_directory_path) else www_directory_path
        self._nginx_www_directory_path = QLineEdit()
        self._nginx_www_directory_path.setText(str(self._configuration.www_directory_path) if self._configuration.www_directory_path is not None else str(self._app.project.configuration.www_directory_path))
        self._nginx_www_directory_path.textChanged.connect(_update_configuration_www_directory_path)
        www_directory_path_layout = QHBoxLayout()
        www_directory_path_layout.addWidget(self._nginx_www_directory_path)

        def find_www_directory_path() -> None:
            with ExceptionCatcher(self):
                found_www_directory_path = QFileDialog.getExistingDirectory(self, 'Serve your site from...', directory=self._nginx_www_directory_path.text())
                if '' != found_www_directory_path:
                    self._nginx_www_directory_path.setText(found_www_directory_path)
        self._nginx_www_directory_path_find = QPushButton('...')
        self._nginx_www_directory_path_find.released.connect(find_www_directory_path)
        www_directory_path_layout.addWidget(self._nginx_www_directory_path_find)
        layout.addRow('WWW directory', www_directory_path_layout)



"""
Provide functional programming utilities.
"""
from __future__ import annotations

from asyncio import sleep
from inspect import isawaitable
from time import time
from typing import Any, Iterable, Sized, TypeVar, Callable, Iterator, Generic, cast, ParamSpec, Awaitable

T = TypeVar('T')
U = TypeVar('U')


def walk(item: Any, attribute_name: str) -> Iterable[Any]:
    """
    Walk over a graph of objects by following a single attribute.
    """
    child = getattr(item, attribute_name)

    # If the child has the requested attribute, yield it,
    if hasattr(child, attribute_name):
        yield child
        yield from walk(child, attribute_name)

    # Otherwise loop over the children and yield their attributes.
    try:
        child = iter(child)
    except TypeError:
        return
    for child in child:
        yield child
        yield from walk(child, attribute_name)


def slice_to_range(indices: slice, iterable: Sized) -> Iterable[int]:
    """
    Apply a slice to an iterable, and return the corresponding range.
    """
    length = len(iterable)

    if indices.start is None:
        start = 0
    else:
        # Ensure the stop index is within range.
        start = max(-length, min(length, indices.start))

    if indices.stop is None:
        stop = max(0, length)
    else:
        # Ensure the stop index is within range.
        stop = max(-length, min(length, indices.stop))

    if indices.step is None:
        step = 1
    else:
        step = indices.step

    return range(start, stop, step)


class _Result(Generic[T]):
    def __init__(self, value: T | None, _error: BaseException | None = None):
        assert not _error or value is None
        self._value = value
        self._error = _error

    @property
    def value(self) -> T:
        if self._error:
            raise self._error
        return cast(T, self._value)

    def map(self, f: Callable[[T], U]) -> _Result[U]:
        if self._error:
            return cast(_Result[U], self)
        try:
            return _Result(f(self.value))
        except BaseException as e:
            return _Result(None, e)


def filter_suppress(raising_filter: Callable[[T], Any], exception_type: type[BaseException], items: Iterable[T]) -> Iterator[T]:
    """
    Filter values, skipping those for which the application of `raising_filter` raises errors.
    """
    for item in items:
        try:
            raising_filter(item)
            yield item
        except exception_type:
            continue


_DoFReturnT = TypeVar('_DoFReturnT')
_DoFP = ParamSpec('_DoFP')


class Do(Generic[_DoFP, _DoFReturnT]):
    def __init__(
        self,
        f: Callable[_DoFP, _DoFReturnT | Awaitable[_DoFReturnT]],
        *args: _DoFP.args,
        **kwargs: _DoFP.kwargs,
    ):
        self._f = f
        self._args = args
        self._kwargs = kwargs

    async def until(
        self,
        *conditions: Callable[[_DoFReturnT], None | bool | Awaitable[None | bool]],
        retries: int = 5,
        timeout: int = 300,
        interval: int | float = 0.1,
    ) -> _DoFReturnT:
        start_time = time()
        while True:
            retries -= 1
            try:
                f_result_or_coroutine = self._f(*self._args, **self._kwargs)
                if isawaitable(f_result_or_coroutine):
                    f_result = await f_result_or_coroutine
                else:
                    f_result = f_result_or_coroutine
                for condition in conditions:
                    condition_result_or_coroutine = condition(f_result)
                    if isawaitable(condition_result_or_coroutine):
                        condition_result = await condition_result_or_coroutine
                    else:
                        condition_result = cast(None | bool, condition_result_or_coroutine)
                    if condition_result is False:
                        raise RuntimeError(f'Condition {condition} was not met for {f_result}.')
            except BaseException:
                if retries == 0:
                    raise
                if time() - start_time > timeout:
                    raise
                await sleep(interval)
            else:
                return f_result



"""
Provide the Ancestry loading API.
"""
import logging

from betty.app import App
from betty.warnings import deprecated


@deprecated('This function is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. Instead, use `logging.getLogger()`.')
def getLogger() -> logging.Logger:
    """
    Get the ancestry loading logger.
    """
    return logging.getLogger(__name__)


class Loader:
    async def load(self) -> None:
        raise NotImplementedError(repr(self))


class PostLoader:
    async def post_load(self) -> None:
        raise NotImplementedError(repr(self))


async def load(app: App) -> None:
    """
    Load an ancestry.
    """
    await app.dispatcher.dispatch(Loader)()
    await app.dispatcher.dispatch(PostLoader)()



"""Integrate Betty with `Leaflet.js <https://leafletjs.com/>`_."""
from __future__ import annotations

import asyncio
import logging
from contextlib import suppress
from pathlib import Path
from shutil import copy2, copytree

from aiofiles.os import makedirs

from betty.app.extension import Extension, UserFacingExtension
from betty.extension.npm import _Npm, _NpmBuilder, npm, _NpmBuilderCacheScope
from betty.generate import Generator, GenerationContext
from betty.html import CssProvider, JsProvider
from betty.locale import Str


class _Maps(UserFacingExtension, CssProvider, JsProvider, Generator, _NpmBuilder):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await npm(('run', 'webpack'), cwd=working_directory_path)
        await self._copy_npm_build(working_directory_path / 'webpack-build', assets_directory_path)
        logging.getLogger(__name__).info(self._app.localizer._('Built the interactive maps.'))

    async def _copy_npm_build(self, source_directory_path: Path, destination_directory_path: Path) -> None:
        await makedirs(destination_directory_path, exist_ok=True)
        await asyncio.to_thread(copy2, source_directory_path / 'maps.css', destination_directory_path / 'maps.css')
        await asyncio.to_thread(copy2, source_directory_path / 'maps.js', destination_directory_path / 'maps.js')
        with suppress(FileNotFoundError):
            await asyncio.to_thread(copytree, source_directory_path / 'images', destination_directory_path / 'images')

    @classmethod
    def npm_cache_scope(cls) -> _NpmBuilderCacheScope:
        return _NpmBuilderCacheScope.BETTY

    async def generate(self, job_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await makedirs(self.app.project.configuration.www_directory_path, exist_ok=True)
        await self._copy_npm_build(assets_directory_path, self.app.project.configuration.www_directory_path)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @property
    def public_css_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('maps.css'),
        ]

    @property
    def public_js_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('maps.js'),
        ]

    @classmethod
    def label(cls) -> Str:
        return Str._('Maps')

    @classmethod
    def description(cls) -> Str:
        return Str._('Display lists of places as interactive maps using <a href="https://leafletjs.com/">Leaflet</a>.')



"""
Provide logging utilities.
"""
import sys
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET, StreamHandler, LogRecord


class CliHandler(
    StreamHandler,  # type: ignore[type-arg]
):
    COLOR_LEVELS = {
        CRITICAL: 91,
        ERROR: 91,
        WARNING: 93,
        INFO: 92,
        DEBUG: 97,
        NOTSET: 97,
    }

    def __init__(self):
        StreamHandler.__init__(self, sys.stderr)

    def format(self, record: LogRecord) -> str:
        s = StreamHandler.format(self, record)
        for level, color in self.COLOR_LEVELS.items():
            if record.levelno >= level:
                return self._color(s, color)
        return self._color(s, self.COLOR_LEVELS[NOTSET])

    def _color(self, s: str, color: int) -> str:
        return '\033[%dm%s\033[0m' % (color, s)



"""
Provide a URL generation API.
"""
from __future__ import annotations

from contextlib import suppress
from typing import Any
from urllib.parse import quote

from betty.app import App
from betty.locale import negotiate_locale, Localey, to_locale
from betty.model import get_entity_type_name, Entity
from betty.project import ProjectConfiguration
from betty.string import camel_case_to_kebab_case


class LocalizedUrlGenerator:
    def generate(self, resource: Any, media_type: str, absolute: bool = False, locale: Localey | None = None) -> str:
        raise NotImplementedError(repr(self))


class StaticUrlGenerator:
    def generate(self, resource: Any, absolute: bool = False, ) -> str:
        raise NotImplementedError(repr(self))


class LocalizedPathUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, app: App):
        self._app = app

    def generate(self, resource: Any, media_type: str, absolute: bool = False, locale: Localey | None = None) -> str:
        return _generate_from_path(
            self._app.project.configuration,
            resource,
            absolute,
            self._app.localizer.locale if locale is None else locale,
        )


class StaticPathUrlGenerator(StaticUrlGenerator):
    def __init__(self, configuration: ProjectConfiguration):
        self._configuration = configuration

    def generate(self, resource: Any, absolute: bool = False, ) -> str:
        return _generate_from_path(self._configuration, resource, absolute)


class _EntityUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, app: App, entity_type: type[Entity]):
        self._app = app
        self._entity_type = entity_type
        self._pattern = f'{camel_case_to_kebab_case(get_entity_type_name(entity_type))}/{{entity_id}}/index.{{extension}}'

    def generate(self, resource: Entity, media_type: str, absolute: bool = False, locale: Localey | None = None) -> str:
        if not isinstance(resource, self._entity_type):
            raise ValueError('%s is not a %s' % (type(resource), self._entity_type))

        if 'text/html' == media_type:
            extension = 'html'
            if locale is None:
                locale = self._app.localizer.locale
        elif 'application/json' == media_type:
            extension = 'json'
            locale = None
        else:
            raise ValueError(f'Unknown entity media type "{media_type}".')

        return _generate_from_path(
            self._app.project.configuration,
            self._pattern.format(
                entity_id=quote(resource.id),
                extension=extension,
            ),
            absolute,
            locale,
        )


class AppUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, app: App):
        self._generators = [
            *(
                _EntityUrlGenerator(app, entity_type)
                for entity_type in app.entity_types
            ),
            LocalizedPathUrlGenerator(app),
        ]

    def generate(self, resource: Any, media_type: str, absolute: bool = False, locale: Localey | None = None) -> str:
        for generator in self._generators:
            with suppress(ValueError):
                return generator.generate(resource, media_type, absolute, locale)
        raise ValueError('No URL generator found for %s.' % (
            resource if isinstance(resource, str) else type(resource)))


def _generate_from_path(configuration: ProjectConfiguration, path: str, absolute: bool = False, localey: Localey | None = None) -> str:
    url = configuration.base_url if absolute else ''
    url += '/'
    if configuration.root_path:
        url += configuration.root_path + '/'
    if localey and configuration.locales.multilingual:
        locale = to_locale(localey)
        try:
            locale_configuration = configuration.locales[locale]
        except KeyError:
            project_locales = list(configuration.locales)
            try:
                negotiated_locale_data = negotiate_locale(locale, project_locales)
                if negotiated_locale_data is None:
                    raise KeyError
                locale_configuration = configuration.locales[to_locale(negotiated_locale_data)]
            except KeyError:
                raise ValueError(f'Cannot generate URLs in "{locale}", because it cannot be resolved to any of the enabled project locales: {", ".join(project_locales)}')
        url += locale_configuration.alias + '/'
    url += path.strip('/')
    if configuration.clean_urls and url.endswith('/index.html'):
        url = url[:-10]
    return url.rstrip('/')



"""
Provide Cotton Candy's Graphical User Interface.
"""
from __future__ import annotations

from typing import Any

from PyQt6.QtCore import QRect
from PyQt6.QtGui import QPainter, QBrush, QColor, QPaintEvent
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QPushButton, QColorDialog, QHBoxLayout, QFormLayout, QWidget

from betty.app import App
from betty.extension import CottonCandy
from betty.extension.cotton_candy import _ColorConfiguration, CottonCandyConfiguration
from betty.gui.locale import LocalizedObject
from betty.gui.model import EntityReferenceSequenceCollector
from betty.locale import Str, Localizable


class _ColorConfigurationSwatch(LocalizedObject, QWidget):
    def __init__(self, app: App, color: _ColorConfiguration, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._color = color
        self._color.on_change(self.repaint)
        self.setFixedHeight(24)
        self.setFixedWidth(24)

    def paintEvent(self, a0: QPaintEvent | None) -> None:
        painter = QPainter(self)
        swatch = QRect(self.rect())
        painter.fillRect(swatch, QBrush(QColor.fromString(self._color.hex)))
        painter.drawRect(swatch)


class _ColorConfigurationWidget(LocalizedObject, QWidget):
    def __init__(self, app: App, color: _ColorConfiguration, color_default: str, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._color = color
        self._color_default = color_default
        self._layout = QHBoxLayout()
        self.setLayout(self._layout)

        self._label = QLabel()
        self._layout.addWidget(self._label)
        self._swatch = _ColorConfigurationSwatch(app, self._color)
        self._layout.addWidget(self._swatch)
        self._configure = QPushButton()
        self._layout.addWidget(self._configure)

        def _configure() -> None:
            qcolor = QColorDialog.getColor(
                initial=QColor.fromString(self._color.hex)
            )
            if qcolor.isValid():
                self._color.hex = qcolor.name()
        self._configure.clicked.connect(_configure)
        self._reset = QPushButton()
        self._layout.addWidget(self._reset)

        def _reset() -> None:
            self._color.hex = self._color_default

        self._reset.clicked.connect(_reset)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._configure.setText(self._app.localizer._('Configure'))
        self._reset.setText(self._app.localizer._('Reset'))


class _ColorConfigurationsWidget(LocalizedObject, QWidget):
    def __init__(self, app: App, colors: list[tuple[_ColorConfiguration, Localizable, str]], *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._colors = colors
        self._color_configurations = []
        self._color_labels = []
        self._layout = QFormLayout()
        self.setLayout(self._layout)

        for color, __, color_default in colors:
            color_label = QLabel()
            self._color_labels.append(color_label)
            color_widget = _ColorConfigurationWidget(app, color, color_default)
            self._color_configurations.append(color_widget)
            self._layout.addRow(color_label, color_widget)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        for i, (__, color_label, ___) in enumerate(self._colors):
            self._color_labels[i].setText(color_label.localize(self._app.localizer))


class _CottonCandyGuiWidget(LocalizedObject, QWidget):
    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._app = app

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._color_configurations_widget = _ColorConfigurationsWidget(app, [
            (
                app.extensions[CottonCandy].configuration.primary_inactive_color,
                Str._('Primary color (inactive)'),
                CottonCandyConfiguration.DEFAULT_PRIMARY_INACTIVE_COLOR,
            ),
            (
                app.extensions[CottonCandy].configuration.primary_active_color,
                Str._('Primary color (active)'),
                CottonCandyConfiguration.DEFAULT_PRIMARY_ACTIVE_COLOR,
            ),
            (
                app.extensions[CottonCandy].configuration.link_inactive_color,
                Str._('Link color (inactive)'),
                CottonCandyConfiguration.DEFAULT_LINK_INACTIVE_COLOR,
            ),
            (
                app.extensions[CottonCandy].configuration.link_active_color,
                Str._('Link color (active)'),
                CottonCandyConfiguration.DEFAULT_LINK_ACTIVE_COLOR,
            ),
        ])
        self._layout.addWidget(self._color_configurations_widget)

        self._featured_entities_label = QLabel()
        self._layout.addWidget(self._featured_entities_label)
        self._featured_entities_entity_references_collector = EntityReferenceSequenceCollector(
            self._app,
            self._app.extensions[CottonCandy].configuration.featured_entities,
            Str._('Featured entities'),
            Str._("These entities are featured on your site's front page."),
        )
        self._layout.addWidget(self._featured_entities_entity_references_collector)



"""
Provide class utilities.
"""
from typing import Any


def repr_instance(instance: object, **attributes: Any) -> str:
    """
    Build a representation of an instance.
    """
    return '<{}.{} object at {}; {}>'.format(
        instance.__class__.__module__,
        instance.__class__.__name__,
        hex(id(instance)),
        (' ' + ', '.join(map(lambda x: f'{x[0]}={x[1]}', attributes.items()))).rstrip(),
    )



"""
Provide an API to determine if information should be kept private.
"""
from __future__ import annotations

import logging
from contextlib import suppress
from datetime import datetime
from typing import Iterator, TypeAlias, Any

from betty.functools import walk
from betty.locale import DateRange, Date, Localizer
from betty.model import Entity
from betty.model.ancestry import Person, Event, HasFiles, HasCitations, HasNotes, Source, \
    Presence, Privacy, HasPrivacy, Subject, Place
from betty.model.event_type import EndOfLifeEventType

Expirable: TypeAlias = Person | Event | Date | None


class Privatizer:
    def __init__(
        self,
        lifetime_threshold: int,
        *,
        localizer: Localizer,
    ):
        self._lifetime_threshold = lifetime_threshold
        self._localizer = localizer
        self._seen: list[HasPrivacy] = []

    def privatize(self, subject: HasPrivacy) -> None:
        if subject.privacy is Privacy.PUBLIC:
            return

        if isinstance(subject, Person):
            self._determine_person_privacy(subject)

        if isinstance(subject, Place):
            self._determine_place_privacy(subject)

        if subject.privacy is not Privacy.PRIVATE:
            return

        if subject in self._seen:
            return
        self._seen.append(subject)

        if isinstance(subject, Person):
            self._privatize_person(subject)

        if isinstance(subject, Presence):
            self._privatize_presence(subject)

        if isinstance(subject, Event):
            self._privatize_event(subject)

        if isinstance(subject, Place):
            self._privatize_place(subject)

        if isinstance(subject, Source):
            self._privatize_source(subject)

        if isinstance(subject, HasCitations):
            self._privatize_has_citations(subject)

        if isinstance(subject, HasFiles):
            self._privatize_has_files(subject)

        if isinstance(subject, HasNotes):
            self._privatize_has_notes(subject)

    def _privatize_person(self, person: Person) -> None:
        if not person.private:
            return

        for person_name in person.names:
            self._mark_private(person_name, person)
            self.privatize(person_name)
        for presence in person.presences:
            self._mark_private(presence, person)
            self.privatize(presence)

    def _privatize_presence(self, presence: Presence) -> None:
        if not presence.private:
            return

        if presence.event is not None and isinstance(presence.role, Subject):
            self._mark_private(presence.event, presence)
            self.privatize(presence.event)
        if presence.person is not None:
            self._mark_private(presence.person, presence)
            self.privatize(presence.person)

    def _privatize_event(self, event: Event) -> None:
        if not event.private:
            return

        for presence in event.presences:
            self._mark_private(presence, event)
            self.privatize(presence)
        if event.place:
            self.privatize(event.place)

    def _privatize_place(self, place: Place) -> None:
        if not place.private:
            return

        for enclosure in place.encloses:
            if not enclosure.encloses:
                continue
            self._mark_private(enclosure.encloses, place)
            self.privatize(enclosure.encloses)
        for enclosure in place.enclosed_by:
            if not enclosure.enclosed_by:
                continue
            self.privatize(enclosure.enclosed_by)

    def _privatize_has_citations(self, has_citations: HasCitations & HasPrivacy) -> None:
        if not has_citations.private:
            return

        for citation in has_citations.citations:
            self._mark_private(citation, has_citations)
            self.privatize(citation)

    def _privatize_source(self, source: Source) -> None:
        if not source.private:
            return

        for contained_source in source.contains:
            self._mark_private(contained_source, source)
            self.privatize(contained_source)
        for citation in source.citations:
            self._mark_private(citation, source)
            self.privatize(citation)

    def _privatize_has_files(self, has_files: HasFiles & HasPrivacy) -> None:
        if not has_files.private:
            return

        for file in has_files.files:
            self._mark_private(file, has_files)
            self.privatize(file)

    def _privatize_has_notes(self, has_notes: HasNotes & HasPrivacy) -> None:
        if not has_notes.private:
            return

        for note in has_notes.notes:
            self._mark_private(note, has_notes)
            self.privatize(note)

    def _ancestors_by_generation(self, person: Person, generations_ago: int = 1) -> Iterator[tuple[Person, int]]:
        for parent in person.parents:
            yield parent, generations_ago
            yield from self._ancestors_by_generation(parent, generations_ago + 1)

    def _determine_person_privacy(self, person: Person) -> None:
        # Do not change existing explicit privacy declarations.
        if person.privacy is not Privacy.UNDETERMINED:
            return

        # A dead person is not private, regardless of when they died.
        for presence in person.presences:
            if presence.event and issubclass(presence.event.event_type, EndOfLifeEventType):
                if presence.event.date is None:
                    person.public = True
                    return
                if self.has_expired(presence.event, 0):
                    person.public = True
                    return

        if self.has_expired(person, 1):
            person.public = True
            return

        for ancestor, generations_ago in self._ancestors_by_generation(person):
            if self.has_expired(ancestor, generations_ago + 1):
                person.public = True
                return

        # If any descendant has any expired event, the person is considered not private.
        for descendant in walk(person, 'children'):
            if self.has_expired(descendant, 1):
                person.public = True
                return

        person.private = True
        logging.getLogger(__name__).debug(self._localizer._('Privatized person {privatized_person_id} ({privatized_person}) because they are likely still alive.').format(
            privatized_person_id=person.id,
            privatized_person=person.label.localize(self._localizer),
        ))

    def _determine_place_privacy(self, place: Place) -> None:
        # Do not change existing explicit privacy declarations.
        if place.privacy is not Privacy.UNDETERMINED:
            return

        # If there are non-private events, we will not privatize the place.
        for event in place.events:
            if not event.private:
                return

        # If there are non-private enclosed places, we will not privatize the place.
        for enclosure in place.encloses:
            if not enclosure.encloses:
                continue
            if not enclosure.encloses.private:
                return

        place.private = True
        logging.getLogger(__name__).debug(self._localizer._('Privatized place {privatized_place_id} ({privatized_place}) because it is not associated with any public information.').format(
            privatized_place_id=place.id,
            privatized_place=place.label.localize(self._localizer),
        ))

    def has_expired(
        self,
        subject: Expirable,
        generations_ago: int = 0,
    ) -> bool:
        if isinstance(subject, Person):
            return self._person_has_expired(subject, generations_ago)

        if isinstance(subject, Event):
            return self._event_has_expired(subject, generations_ago)

        if isinstance(subject, Date):
            return self._date_has_expired(subject, generations_ago)

        return False

    def _person_has_expired(self, person: Person, generations_ago: int) -> bool:
        for presence in person.presences:
            if presence.event is not None and self._event_has_expired(presence.event, generations_ago):
                return True
        return False

    def _event_has_expired(self, event: Event, generations_ago: int) -> bool:
        date = event.date

        if isinstance(date, DateRange):
            # We can only determine event expiration with certainty if we have an end date to work with. Someone born in
            # 2000 can have a valid birth event with a start date of 1800, which does nothing to help us determine
            # expiration.
            date = date.end

        return self.has_expired(date, generations_ago)

    def _date_has_expired(
        self,
        date: Date,
        generations_ago: int,
    ) -> bool:
        if not date.comparable:
            return False

        return date <= Date(datetime.now().year - self._lifetime_threshold * generations_ago, datetime.now().month, datetime.now().day)

    def _mark_private(self, target: HasPrivacy, reason: Any) -> None:
        # Do not change existing explicit privacy declarations.
        if target.own_privacy is not Privacy.UNDETERMINED:
            return

        target.private = True
        with suppress(ValueError):
            self._seen.remove(target)

        if isinstance(target, Entity) and isinstance(reason, Entity):
            logging.getLogger(__name__).debug(self._localizer._('Privatized {privatized_entity_type} {privatized_entity_id} ({privatized_entity}) because of {reason_entity_type} {reason_entity_id} ({reason_entity}).').format(
                privatized_entity_type=target.entity_type_label().localize(self._localizer),
                privatized_entity_id=target.id,
                privatized_entity=target.label.localize(self._localizer),
                reason_entity_type=reason.entity_type_label().localize(self._localizer),
                reason_entity_id=reason.id,
                reason_entity=reason.label.localize(self._localizer),
            ))



"""
Provide Betty's ancestry event types.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from betty.locale import Str, DEFAULT_LOCALIZER

if TYPE_CHECKING:
    from betty.model.ancestry import Person


class EventTypeProvider:
    @property
    def entity_types(self) -> set[type[EventType]]:
        raise NotImplementedError(repr(self))


class EventType:
    def __new__(cls):
        raise RuntimeError('Event types cannot be instantiated.')

    @classmethod
    def name(cls) -> str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return set()

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return set()


class UnknownEventType(EventType):
    @classmethod
    def name(cls) -> str:
        return 'unknown'

    @classmethod
    def label(cls) -> Str:
        return Str._('Unknown')


class DerivableEventType(EventType):
    pass


class CreatableDerivableEventType(DerivableEventType):
    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        return True


class PreBirthEventType(EventType):
    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Birth}


class StartOfLifeEventType(EventType):
    pass


class DuringLifeEventType(EventType):
    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Birth}

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Death}


class EndOfLifeEventType(EventType):
    pass


class PostDeathEventType(EventType):
    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Death}


class Birth(CreatableDerivableEventType, StartOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'birth'

    @classmethod
    def label(cls) -> Str:
        return Str._('Birth')

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {DuringLifeEventType}


class Baptism(DuringLifeEventType, StartOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'baptism'

    @classmethod
    def label(cls) -> Str:
        return Str._('Baptism')


class Adoption(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'adoption'

    @classmethod
    def label(cls) -> Str:
        return Str._('Adoption')


class Death(CreatableDerivableEventType, EndOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'death'

    @classmethod
    def label(cls) -> Str:
        return Str._('Death')

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {DuringLifeEventType}

    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        from betty.privatizer import Privatizer

        return Privatizer(lifetime_threshold, localizer=DEFAULT_LOCALIZER).has_expired(person, 1)


class FinalDispositionEventType(PostDeathEventType, DerivableEventType, EndOfLifeEventType):
    pass


class Funeral(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return 'funeral'

    @classmethod
    def label(cls) -> Str:
        return Str._('Funeral')


class Cremation(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return 'cremation'

    @classmethod
    def label(cls) -> Str:
        return Str._('Cremation')


class Burial(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return 'burial'

    @classmethod
    def label(cls) -> Str:
        return Str._('Burial')


class Will(PostDeathEventType):
    @classmethod
    def name(cls) -> str:
        return 'will'

    @classmethod
    def label(cls) -> Str:
        return Str._('Will')


class Engagement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'engagement'

    @classmethod
    def label(cls) -> Str:
        return Str._('Engagement')

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Marriage}


class Marriage(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'marriage'

    @classmethod
    def label(cls) -> Str:
        return Str._('Marriage')


class MarriageAnnouncement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'marriage-announcement'

    @classmethod
    def label(cls) -> Str:
        return Str._('Announcement of marriage')

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Marriage}


class Divorce(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'divorce'

    @classmethod
    def label(cls) -> Str:
        return Str._('Divorce')

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Marriage}


class DivorceAnnouncement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'divorce-announcement'

    @classmethod
    def label(cls) -> Str:
        return Str._('Announcement of divorce')

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Marriage}

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Divorce}


class Residence(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'residence'

    @classmethod
    def label(cls) -> Str:
        return Str._('Residence')


class Immigration(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'immigration'

    @classmethod
    def label(cls) -> Str:
        return Str._('Immigration')


class Emigration(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'emigration'

    @classmethod
    def label(cls) -> Str:
        return Str._('Emigration')


class Occupation(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'occupation'

    @classmethod
    def label(cls) -> Str:
        return Str._('Occupation')


class Retirement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'retirement'

    @classmethod
    def label(cls) -> Str:
        return Str._('Retirement')


class Correspondence(EventType):
    @classmethod
    def name(cls) -> str:
        return 'correspondence'

    @classmethod
    def label(cls) -> Str:
        return Str._('Correspondence')


class Confirmation(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'confirmation'

    @classmethod
    def label(cls) -> Str:
        return Str._('Confirmation')


class Missing(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'missing'

    @classmethod
    def label(cls) -> Str:
        return Str._('Missing')


class Conference(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'conference'

    @classmethod
    def label(cls) -> Str:
        return Str._('Conference')



"""
Provide text widgets for the Graphical User Interface.
"""

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QLabel


class Text(QLabel):
    def __init__(self, text: str | None = None):
        super().__init__()
        if text:
            self.setText(text)
        self.setTextFormat(Qt.TextFormat.RichText)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard | Qt.TextInteractionFlag.LinksAccessibleByMouse | Qt.TextInteractionFlag.TextSelectableByKeyboard | Qt.TextInteractionFlag.TextSelectableByMouse)
        self.setOpenExternalLinks(True)


class Caption(Text):
    def __init__(self, text: str | None = None):
        super().__init__(text)
        font = QFont()
        self.setFont(font)


class Code(Text):
    def __init__(self, text: str | None = None):
        super().__init__(text)
        font = QFont()
        self.setFont(font)

    def setText(self, a0: str | None) -> None:
        super().setText(f'<pre>{a0}</pre>' if a0 else a0)



"""Provide Betty's data model API."""

from __future__ import annotations

import builtins
import functools
import weakref
from collections import defaultdict
from contextlib import contextmanager
from reprlib import recursive_repr
from typing import TypeVar, Generic, Iterable, Any, overload, cast, Iterator, Callable, Self, TypeAlias, TYPE_CHECKING
from uuid import uuid4

from betty.classtools import repr_instance
from betty.importlib import import_any, fully_qualified_type_name
from betty.json.linked_data import LinkedDataDumpable, add_json_ld
from betty.json.schema import ref_json_schema
from betty.locale import Str
from betty.serde.dump import DictDump, Dump
from betty.string import camel_case_to_kebab_case, upper_camel_case_to_lower_camel_case

if TYPE_CHECKING:
    from betty.app import App


T = TypeVar('T')


class GeneratedEntityId(str):
    """
    Generate a unique entity ID.

    Entities must have IDs for identification. However, not all entities can be provided with an ID that exists in the
    original data set (such as a third-party family tree loaded into Betty), so IDs can be generated.
    """

    def __new__(cls, entity_id: str | None = None):
        return super().__new__(cls, entity_id or str(uuid4()))


class Entity(LinkedDataDumpable):
    def __init__(
        self,
        id: str | None = None,
        *args: Any,
        **kwargs: Any,
    ):
        self._id = GeneratedEntityId() if id is None else id
        super().__init__(*args, **kwargs)

    def __hash__(self) -> int:
        return hash(self.ancestry_id)

    @classmethod
    def entity_type_label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, id=self._id)

    @property
    def type(self) -> builtins.type[Self]:
        return self.__class__

    @property
    def id(self) -> str:
        return self._id

    @property
    def ancestry_id(self) -> tuple[builtins.type[Self], str]:
        return self.type, self.id

    @property
    def label(self) -> Str:
        return Str._(
            '{entity_type} {entity_id}',
            entity_type=self.entity_type_label(),
            entity_id=self.id,
        )

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)

        entity_type_name = get_entity_type_name(self.type)
        dump['$schema'] = app.static_url_generator.generate(
            f'schema.json#/definitions/entity/{upper_camel_case_to_lower_camel_case(entity_type_name)}',
            absolute=True,
        )

        if not isinstance(self.id, GeneratedEntityId):
            dump['@id'] = app.static_url_generator.generate(
                f'/{camel_case_to_kebab_case(entity_type_name)}/{self.id}/index.json',
                absolute=True,
            )
            dump['id'] = self.id

        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        schema['type'] = 'object'
        schema['properties'] = {
            '$schema': ref_json_schema(schema),
            'id': {
                'type': 'string',
            },
        }
        schema['required'] = [
            '$schema',
        ]
        schema['additionalProperties'] = False
        add_json_ld(schema)
        return schema


AncestryEntityId: TypeAlias = tuple[type[Entity], str]


class UserFacingEntity:
    pass


class EntityTypeProvider:
    async def entity_types(self) -> set[type[Entity]]:
        raise NotImplementedError(repr(self))


EntityT = TypeVar('EntityT', bound=Entity)
EntityU = TypeVar('EntityU', bound=Entity)
TargetT = TypeVar('TargetT')
OwnerT = TypeVar('OwnerT')
AssociateT = TypeVar('AssociateT')
AssociateU = TypeVar('AssociateU')
LeftAssociateT = TypeVar('LeftAssociateT')
RightAssociateT = TypeVar('RightAssociateT')


def get_entity_type_name(entity_type_definition: type[Entity] | Entity) -> str:
    """
    Get the entity type name for an entity or entity type.
    """
    if isinstance(entity_type_definition, Entity):
        entity_type = entity_type_definition.type
    else:
        entity_type = entity_type_definition

    if entity_type.__module__.startswith('betty.model.ancestry'):
        return entity_type.__name__
    return f'{entity_type.__module__}.{entity_type.__name__}'


def get_entity_type(entity_type_name: str) -> type[Entity]:
    """
    Get the entity type for an entity type name.
    """
    try:
        return import_any(entity_type_name)  # type: ignore[no-any-return]
    except ImportError:
        try:
            return import_any(f'betty.model.ancestry.{entity_type_name}')  # type: ignore[no-any-return]
        except ImportError:
            raise EntityTypeImportError(entity_type_name) from None


class EntityTypeError(ValueError):
    pass


class EntityTypeImportError(EntityTypeError, ImportError):
    """
    Raised when an alleged entity type cannot be imported.
    """

    def __init__(self, entity_type_name: str):
        super().__init__(f'Cannot find and import an entity with name "{entity_type_name}".')


class EntityTypeInvalidError(EntityTypeError, ImportError):
    """
    Raised for types that are not valid entity types.
    """

    def __init__(self, entity_type: type):
        super().__init__(f'{entity_type.__module__}.{entity_type.__name__} is not an entity type class. Entity types must extend {Entity.__module__}.{Entity.__name__} directly.')


class EntityCollection(Generic[TargetT]):
    def __init__(self):
        super().__init__()

    def _on_add(self, *entities: TargetT & Entity) -> None:
        pass

    def _on_remove(self, *entities: TargetT & Entity) -> None:
        pass

    @property
    def view(self) -> list[TargetT & Entity]:
        return [*self]

    def add(self, *entities: TargetT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def remove(self, *entities: TargetT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def replace(self, *entities: TargetT & Entity) -> None:
        self.remove(*(
            entity
            for entity
            in self
            if entity not in entities
        ))
        self.add(*entities)

    def clear(self) -> None:
        raise NotImplementedError(repr(self))

    def __iter__(self) -> Iterator[TargetT & Entity]:
        raise NotImplementedError(repr(self))

    def __len__(self) -> int:
        raise NotImplementedError(repr(self))

    @overload
    def __getitem__(self, index: int) -> TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> list[TargetT & Entity]:
        pass

    def __getitem__(self, key: int | slice) -> TargetT & Entity | list[TargetT & Entity]:
        raise NotImplementedError(repr(self))

    def __delitem__(self, key: TargetT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def __contains__(self, value: Any) -> bool:
        raise NotImplementedError(repr(self))

    def _known(self, *entities: TargetT & Entity) -> Iterable[TargetT & Entity]:
        seen = []
        for entity in entities:
            if entity in self and entity not in seen:
                yield entity
                seen.append(entity)

    def _unknown(self, *entities: TargetT & Entity) -> Iterable[TargetT & Entity]:
        seen = []
        for entity in entities:
            if entity not in self and entity not in seen:
                yield entity
                seen.append(entity)


EntityCollectionT = TypeVar('EntityCollectionT', bound=EntityCollection[EntityT])


class _EntityTypeAssociation(Generic[OwnerT, AssociateT]):
    def __init__(
        self,
        owner_type: type[OwnerT],
        owner_attr_name: str,
        associate_type_name: str,
    ):
        self._owner_type = owner_type
        self._owner_attr_name = owner_attr_name
        self._owner_private_attr_name = f'_{owner_attr_name}'
        self._associate_type_name = associate_type_name
        self._associate_type: type[AssociateT] | None = None

    def __hash__(self) -> int:
        return hash((
            self._owner_type,
            self._owner_attr_name,
            self._associate_type_name,
        ))

    def __repr__(self) -> str:
        return repr_instance(
            self,
            owner_type=self._owner_type,
            owner_attr_name=self._owner_attr_name,
            associate_type_name=self._associate_type_name,
        )

    @property
    def owner_type(self) -> type[OwnerT]:
        return self._owner_type

    @property
    def owner_attr_name(self) -> str:
        return self._owner_attr_name

    @property
    def associate_type(self) -> type[AssociateT]:
        if self._associate_type is None:
            self._associate_type = import_any(self._associate_type_name)
        return self._associate_type

    def register(  # type: ignore[misc]
        self: ToAny[OwnerT, AssociateT],
    ) -> None:
        EntityTypeAssociationRegistry._register(self)

        original_init = self._owner_type.__init__

        @functools.wraps(original_init)
        def _init(owner: OwnerT & Entity, *args: Any, **kwargs: Any) -> None:
            self.initialize(owner)
            original_init(owner, *args, **kwargs)
        self._owner_type.__init__ = _init  # type: ignore[assignment, method-assign]

    def initialize(self, owner: OwnerT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def finalize(self, owner: OwnerT & Entity) -> None:
        self.delete(owner)
        delattr(owner, self._owner_private_attr_name)

    def delete(self, owner: OwnerT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def associate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        raise NotImplementedError(repr(self))

    def disassociate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        raise NotImplementedError(repr(self))


class BidirectionalEntityTypeAssociation(Generic[OwnerT, AssociateT], _EntityTypeAssociation[OwnerT, AssociateT]):
    def __init__(
        self,
        owner_type: type[OwnerT],
        owner_attr_name: str,
        associate_type_name: str,
        associate_attr_name: str,
    ):
        super().__init__(
            owner_type,
            owner_attr_name,
            associate_type_name,
        )
        self._associate_attr_name = associate_attr_name

    def __hash__(self) -> int:
        return hash((
            self._owner_type,
            self._owner_attr_name,
            self._associate_type_name,
            self._associate_attr_name,
        ))

    def __repr__(self) -> str:
        return repr_instance(
            self,
            owner_type=self._owner_type,
            owner_attr_name=self._owner_attr_name,
            associate_type_name=self._associate_type_name,
            associate_attr_name=self._associate_attr_name,
        )

    @property
    def associate_attr_name(self) -> str:
        return self._associate_attr_name

    def inverse(self) -> BidirectionalEntityTypeAssociation[AssociateT, OwnerT]:
        association = EntityTypeAssociationRegistry.get_association(self.associate_type, self.associate_attr_name)
        assert isinstance(association, BidirectionalEntityTypeAssociation)
        return association


class ToOneEntityTypeAssociation(Generic[OwnerT, AssociateT], _EntityTypeAssociation[OwnerT, AssociateT]):
    def register(self) -> None:
        super().register()
        setattr(
            self.owner_type,
            self.owner_attr_name,
            property(
                self.get,
                self.set,
                self.delete,
            ),
        )

    def initialize(self, owner: OwnerT & Entity) -> None:
        setattr(owner, self._owner_private_attr_name, None)

    def get(self, owner: OwnerT & Entity) -> AssociateT & Entity | None:
        return getattr(owner, self._owner_private_attr_name)  # type: ignore[no-any-return]

    def set(self, owner: OwnerT & Entity, associate: AssociateT & Entity | None) -> None:
        setattr(owner, self._owner_private_attr_name, associate)

    def delete(self, owner: OwnerT & Entity) -> None:
        self.set(owner, None)

    def associate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        self.set(owner, associate)

    def disassociate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        if associate == self.get(owner):
            self.delete(owner)


class ToManyEntityTypeAssociation(Generic[OwnerT, AssociateT], _EntityTypeAssociation[OwnerT, AssociateT]):
    def register(self) -> None:
        super().register()
        setattr(
            self.owner_type,
            self.owner_attr_name,
            property(
                self.get,
                self.set,
                self.delete,
            ),
        )

    def get(self, owner: OwnerT & Entity) -> EntityCollection[AssociateT & Entity]:
        return cast(EntityCollection['AssociateT & Entity'], getattr(owner, self._owner_private_attr_name))

    def set(self, owner: OwnerT & Entity, entities: Iterable[AssociateT & Entity]) -> None:
        self.get(owner).replace(*entities)

    def delete(self, owner: OwnerT & Entity) -> None:
        self.get(owner).clear()

    def associate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        self.get(owner).add(associate)

    def disassociate(self, owner: OwnerT & Entity, associate: AssociateT & Entity) -> None:
        self.get(owner).remove(associate)


class BidirectionalToOneEntityTypeAssociation(
    Generic[OwnerT, AssociateT],
    ToOneEntityTypeAssociation[OwnerT, AssociateT],
    BidirectionalEntityTypeAssociation[OwnerT, AssociateT]
):
    def set(self, owner: OwnerT & Entity, associate: AssociateT & Entity | None) -> None:
        previous_associate = self.get(owner)
        if previous_associate == associate:
            return
        super().set(owner, associate)
        if previous_associate is not None:
            self.inverse().disassociate(previous_associate, owner)
        if associate is not None:
            self.inverse().associate(associate, owner)


class BidirectionalToManyEntityTypeAssociation(
    Generic[OwnerT, AssociateT],
    ToManyEntityTypeAssociation[OwnerT, AssociateT],
    BidirectionalEntityTypeAssociation[OwnerT, AssociateT],
):
    def initialize(self, owner: OwnerT & Entity) -> None:
        setattr(
            owner,
            self._owner_private_attr_name,
            _BidirectionalAssociateCollection(
                owner,
                self,
            )
        )


class ToOne(Generic[OwnerT, AssociateT], ToOneEntityTypeAssociation[OwnerT, AssociateT]):
    pass


class OneToOne(Generic[OwnerT, AssociateT], BidirectionalToOneEntityTypeAssociation[OwnerT, AssociateT]):
    pass


class ManyToOne(Generic[OwnerT, AssociateT], BidirectionalToOneEntityTypeAssociation[OwnerT, AssociateT]):
    pass


class ToMany(Generic[OwnerT, AssociateT], ToManyEntityTypeAssociation[OwnerT, AssociateT]):
    def initialize(self, owner: OwnerT & Entity) -> None:
        setattr(
            owner,
            self._owner_private_attr_name,
            SingleTypeEntityCollection[AssociateT](self.associate_type)
        )


class OneToMany(Generic[OwnerT, AssociateT], BidirectionalToManyEntityTypeAssociation[OwnerT, AssociateT]):
    pass


class ManyToMany(Generic[OwnerT, AssociateT], BidirectionalToManyEntityTypeAssociation[OwnerT, AssociateT]):
    pass


ToAny: TypeAlias = ToOneEntityTypeAssociation[OwnerT, AssociateT] | ToManyEntityTypeAssociation[OwnerT, AssociateT]


def to_one(
    owner_attr_name: str,
    associate_type_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a unidirectional to-one association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        ToOne(
            owner_type,
            owner_attr_name,
            associate_type_name,
        ).register()
        return owner_type
    return _decorator


def one_to_one(
    owner_attr_name: str,
    associate_type_name: str,
    associate_attr_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a bidirectional one-to-one association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        OneToOne(
            owner_type,
            owner_attr_name,
            associate_type_name,
            associate_attr_name,
        ).register()
        return owner_type
    return _decorator


def many_to_one(
    owner_attr_name: str,
    associate_type_name: str,
    associate_attr_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a bidirectional many-to-one association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        ManyToOne(
            owner_type,
            owner_attr_name,
            associate_type_name,
            associate_attr_name,
        ).register()
        return owner_type
    return _decorator


def to_many(
    owner_attr_name: str,
    associate_type_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a unidirectional to-many association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        ToMany(
            owner_type,
            owner_attr_name,
            associate_type_name,
        ).register()
        return owner_type
    return _decorator


def one_to_many(
    owner_attr_name: str,
    associate_type_name: str,
    associate_attr_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a bidirectional one-to-many association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        OneToMany(
            owner_type,
            owner_attr_name,
            associate_type_name,
            associate_attr_name,
        ).register()
        return owner_type
    return _decorator


def many_to_many(
    owner_attr_name: str,
    associate_type_name: str,
    associate_attr_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a bidirectional many-to-many association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        ManyToMany(
            owner_type,
            owner_attr_name,
            associate_type_name,
            associate_attr_name,
        ).register()
        return owner_type
    return _decorator


def many_to_one_to_many(
    left_associate_type_name: str,
    left_associate_attr_name: str,
    left_owner_attr_name: str,
    right_owner_attr_name: str,
    right_associate_type_name: str,
    right_associate_attr_name: str,
) -> Callable[[type[OwnerT]], type[OwnerT]]:
    """
    Add a bidirectional many-to-one-to-many association to an entity or entity mixin.
    """
    def _decorator(owner_type: type[OwnerT]) -> type[OwnerT]:
        ManyToOne(
            owner_type,
            left_owner_attr_name,
            left_associate_type_name,
            left_associate_attr_name,
        ).register()
        ManyToOne(
            owner_type,
            right_owner_attr_name,
            right_associate_type_name,
            right_associate_attr_name,
        ).register()
        return owner_type
    return _decorator


class EntityTypeAssociationRegistry:
    _associations = set[ToAny[Any, Any]]()

    @classmethod
    def get_all_associations(cls, owner: type | object) -> set[ToAny[Any, Any]]:
        owner_type = owner if isinstance(owner, type) else type(owner)
        return {
            association
            for association
            in cls._associations
            if association.owner_type in owner_type.__mro__
        }

    @classmethod
    def get_association(cls, owner: type[OwnerT] | OwnerT & Entity, owner_attr_name: str) -> ToAny[OwnerT, Any]:
        for association in cls.get_all_associations(owner):
            if association.owner_attr_name == owner_attr_name:
                return association
        raise ValueError(f'No association exists for {fully_qualified_type_name(owner if isinstance(owner, type) else owner.__class__)}.{owner_attr_name}.')

    @classmethod
    def get_associates(cls, owner: EntityT, association: ToAny[EntityT, AssociateT]) -> Iterable[AssociateT]:
        associates: AssociateT | None | Iterable[AssociateT] = getattr(owner, f'_{association.owner_attr_name}')
        if isinstance(association, ToOneEntityTypeAssociation):
            if associates is None:
                return
            yield cast(AssociateT, associates)
            return
        yield from cast(Iterable[AssociateT], associates)

    @classmethod
    def _register(cls, association: ToAny[Any, Any]) -> None:
        cls._associations.add(association)

    @classmethod
    def initialize(cls, *owners: Entity) -> None:
        for owner in owners:
            for association in cls.get_all_associations(owner):
                association.initialize(owner)

    @classmethod
    def finalize(cls, *owners: Entity) -> None:
        for owner in owners:
            for association in cls.get_all_associations(owner):
                association.finalize(owner)


class SingleTypeEntityCollection(Generic[TargetT], EntityCollection[TargetT]):
    def __init__(
        self,
        target_type: type[TargetT],
    ):
        super().__init__()
        self._entities: list[TargetT & Entity] = []
        self._target_type = target_type

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, target_type=self._target_type, length=len(self))

    def add(self, *entities: TargetT & Entity) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self._entities.append(entity)
        if added_entities:
            self._on_add(*added_entities)

    def remove(self, *entities: TargetT & Entity) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self._entities.remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    def clear(self) -> None:
        self.remove(*self)

    def __iter__(self) -> Iterator[TargetT & Entity]:
        return self._entities.__iter__()

    def __len__(self) -> int:
        return len(self._entities)

    @overload
    def __getitem__(self, index: int) -> TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> list[TargetT & Entity]:
        pass

    @overload
    def __getitem__(self, entity_id: str) -> TargetT & Entity:
        pass

    def __getitem__(self, key: int | slice | str) -> TargetT & Entity | list[TargetT & Entity]:
        if isinstance(key, int):
            return self._getitem_by_index(key)
        if isinstance(key, slice):
            return self._getitem_by_indices(key)
        return self._getitem_by_entity_id(key)

    def _getitem_by_index(self, index: int) -> TargetT & Entity:
        return self._entities[index]

    def _getitem_by_indices(self, indices: slice) -> list[TargetT & Entity]:
        return self.view[indices]

    def _getitem_by_entity_id(self, entity_id: str) -> TargetT & Entity:
        for entity in self._entities:
            if entity_id == entity.id:
                return entity
        raise KeyError(f'Cannot find a {self._target_type} entity with ID "{entity_id}".')

    def __delitem__(self, key: str | TargetT & Entity) -> None:
        if isinstance(key, self._target_type):
            return self._delitem_by_entity(cast('TargetT & Entity', key))
        if isinstance(key, str):
            return self._delitem_by_entity_id(key)
        raise TypeError(f'Cannot find entities by {repr(key)}.')

    def _delitem_by_entity(self, entity: TargetT & Entity) -> None:
        self.remove(entity)

    def _delitem_by_entity_id(self, entity_id: str) -> None:
        for entity in self._entities:
            if entity_id == entity.id:
                self.remove(entity)
                return

    def __contains__(self, value: Any) -> bool:
        if isinstance(value, self._target_type):
            return self._contains_by_entity(cast('TargetT & Entity', value))
        if isinstance(value, str):
            return self._contains_by_entity_id(value)
        return False

    def _contains_by_entity(self, other_entity: TargetT & Entity) -> bool:
        for entity in self._entities:
            if other_entity is entity:
                return True
        return False

    def _contains_by_entity_id(self, entity_id: str) -> bool:
        for entity in self._entities:
            if entity.id == entity_id:
                return True
        return False


SingleTypeEntityCollectionT = TypeVar('SingleTypeEntityCollectionT', bound=SingleTypeEntityCollection[AssociateT])


class MultipleTypesEntityCollection(Generic[TargetT], EntityCollection[TargetT]):
    def __init__(self):
        super().__init__()
        self._collections: dict[type[Entity], SingleTypeEntityCollection[Entity]] = {}

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(
            self,
            entity_types=', '.join(map(get_entity_type_name, self._collections.keys())),
            length=len(self),
        )

    def _get_collection(self, entity_type: type[EntityT]) -> SingleTypeEntityCollection[EntityT]:
        assert issubclass(entity_type, Entity), f'{entity_type} is not an entity type.'
        try:
            return cast(SingleTypeEntityCollection[EntityT], self._collections[entity_type])
        except KeyError:
            self._collections[entity_type] = SingleTypeEntityCollection(entity_type)
            return cast(SingleTypeEntityCollection[EntityT], self._collections[entity_type])

    @overload
    def __getitem__(self, index: int) -> TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> list[TargetT & Entity]:
        pass

    @overload
    def __getitem__(self, entity_type_name: str) -> SingleTypeEntityCollection[Entity]:
        pass

    @overload
    def __getitem__(self, entity_type: type[EntityT]) -> SingleTypeEntityCollection[EntityT]:
        pass

    def __getitem__(
        self,
        key: int | slice | str | type[EntityT],
    ) -> TargetT & Entity | SingleTypeEntityCollection[Entity] | SingleTypeEntityCollection[EntityT] | list[TargetT & Entity]:
        if isinstance(key, int):
            return self._getitem_by_index(key)
        if isinstance(key, slice):
            return self._getitem_by_indices(key)
        if isinstance(key, str):
            return self._getitem_by_entity_type_name(key)
        return self._getitem_by_entity_type(key)

    def _getitem_by_entity_type(self, entity_type: type[EntityT]) -> SingleTypeEntityCollection[EntityT]:
        return self._get_collection(entity_type)

    def _getitem_by_entity_type_name(self, entity_type_name: str) -> SingleTypeEntityCollection[Entity]:
        return self._get_collection(
            get_entity_type(entity_type_name),
        )

    def _getitem_by_index(self, index: int) -> TargetT & Entity:
        return self.view[index]

    def _getitem_by_indices(self, indices: slice) -> list[TargetT & Entity]:
        return self.view[indices]

    def __delitem__(self, key: str | type[TargetT & Entity] | TargetT & Entity) -> None:
        if isinstance(key, type):
            return self._delitem_by_type(
                key,
            )
        if isinstance(key, Entity):
            return self._delitem_by_entity(
                key,  # type: ignore[arg-type]
            )
        return self._delitem_by_entity_type_name(key)

    def _delitem_by_type(self, entity_type: type[TargetT & Entity]) -> None:
        removed_entities = [*self._get_collection(entity_type)]
        self._get_collection(entity_type).clear()
        if removed_entities:
            self._on_remove(*removed_entities)

    def _delitem_by_entity(self, entity: TargetT & Entity) -> None:
        self.remove(entity)

    def _delitem_by_entity_type_name(self, entity_type_name: str) -> None:
        self._delitem_by_type(
            get_entity_type(entity_type_name),  # type: ignore[arg-type]
        )

    def __iter__(self) -> Iterator[TargetT & Entity]:
        for collection in self._collections.values():
            for entity in collection:
                yield cast('TargetT & Entity', entity)

    def __len__(self) -> int:
        return sum(map(len, self._collections.values()))

    def __contains__(self, value: Any) -> bool:
        if isinstance(value, Entity):
            return self._contains_by_entity(value)
        return False

    def _contains_by_entity(self, other_entity: Any) -> bool:
        for entity in self:
            if other_entity is entity:
                return True
        return False

    def add(self, *entities: TargetT & Entity) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self[entity.type].add(entity)
        if added_entities:
            self._on_add(*added_entities)

    def remove(self, *entities: TargetT & Entity) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self[entity.type].remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    def clear(self) -> None:
        removed_entities = (*self,)
        for collection in self._collections.values():
            collection.clear()
        if removed_entities:
            self._on_remove(*removed_entities)


class _BidirectionalAssociateCollection(Generic[AssociateT, OwnerT], SingleTypeEntityCollection[AssociateT]):
    def __init__(
        self,
        owner: OwnerT & Entity,
        association: BidirectionalEntityTypeAssociation[OwnerT, AssociateT],
    ):
        super().__init__(association.associate_type)
        self._association = association
        self.__owner = weakref.ref(owner)

    @property
    def _owner(self) -> OwnerT & Entity:
        owner = self.__owner()
        if owner is None:
            raise RuntimeError("This associate collection's owner no longer exists in memory.")
        return owner

    def _on_add(self, *entities: AssociateT & Entity) -> None:
        super()._on_add(*entities)
        for associate in entities:
            self._association.inverse().associate(associate, self._owner)

    def _on_remove(self, *entities: AssociateT & Entity) -> None:
        super()._on_remove(*entities)
        for associate in entities:
            self._association.inverse().disassociate(associate, self._owner)


class AliasedEntity(Generic[EntityT]):
    def __init__(self, original_entity: EntityT, aliased_entity_id: str | None = None):
        self._entity = original_entity
        self._id = GeneratedEntityId() if aliased_entity_id is None else aliased_entity_id

    def __repr__(self) -> str:
        return repr_instance(self, id=self.id)

    @property
    def type(self) -> builtins.type[Entity]:
        return self._entity.type

    @property
    def id(self) -> str:
        return self._id

    def unalias(self) -> EntityT:
        return self._entity


AliasableEntity: TypeAlias = EntityT | AliasedEntity[EntityT]


def unalias(entity: AliasableEntity[EntityT]) -> EntityT:
    """
    Unalias a potentially aliased entity.
    """
    if isinstance(entity, AliasedEntity):
        return entity.unalias()
    return entity


_EntityGraphBuilderEntities: TypeAlias = dict[type[Entity], dict[str, AliasableEntity[Entity]]]


_EntityGraphBuilderAssociations: TypeAlias = dict[
    type[Entity],  # The owner entity type.
    dict[
        str,  # The owner attribute name.
        dict[
            str,  # The owner ID.
            list[AncestryEntityId]  # The associate IDs.
        ]
    ]
]


class _EntityGraphBuilder:
    def __init__(self):
        self._entities: _EntityGraphBuilderEntities = defaultdict(dict)
        self._associations: _EntityGraphBuilderAssociations = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: list())))
        self._built = False

    def _assert_unbuilt(self) -> None:
        if self._built:
            raise RuntimeError('This entity graph has been built already.')

    def _iter(self) -> Iterator[AliasableEntity[Entity]]:
        for entity_type in self._entities:
            yield from self._entities[entity_type].values()

    def _build_associations(self) -> None:
        for owner_type, owner_attrs in self._associations.items():
            for owner_attr_name, owner_associations in owner_attrs.items():
                association = EntityTypeAssociationRegistry.get_association(owner_type, owner_attr_name)
                for owner_id, associate_ancestry_ids in owner_associations.items():
                    associates = [
                        unalias(self._entities[associate_type][associate_id])
                        for associate_type, associate_id
                        in associate_ancestry_ids
                    ]
                    owner = unalias(self._entities[owner_type][owner_id])
                    if isinstance(association, ToOneEntityTypeAssociation):
                        association.set(
                            owner,
                            associates[0]
                        )
                    else:
                        association.set(
                            owner,
                            associates
                        )

    def build(self) -> Iterator[Entity]:
        self._assert_unbuilt()
        self._built = True

        unaliased_entities = list(map(
            unalias,
            self._iter(),
        ))

        EntityTypeAssociationRegistry.initialize(*unaliased_entities)
        self._build_associations()

        yield from unaliased_entities


class EntityGraphBuilder(_EntityGraphBuilder):
    def add_entity(self, *entities: AliasableEntity[Entity]) -> None:
        self._assert_unbuilt()

        for entity in entities:
            self._entities[entity.type][entity.id] = entity

    def add_association(
        self,
        owner_type: type[Entity],
        owner_id: str,
        owner_attr_name: str,
        associate_type: type[Entity],
        associate_id: str,
    ) -> None:
        self._assert_unbuilt()

        self._associations[owner_type][owner_attr_name][owner_id].append((associate_type, associate_id))


@contextmanager
def record_added(entities: EntityCollection[EntityT]) -> Iterator[MultipleTypesEntityCollection[EntityT]]:
    """
    Record all entities that are added to a collection.
    """
    original = [*entities]
    added = MultipleTypesEntityCollection[EntityT]()
    yield added
    added.add(*[
        entity
        for entity
        in entities
        if entity not in original
    ])



"""
Fetch information from Wikipedia.
"""
from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import mimetypes
import re
from collections.abc import Sequence, MutableSequence
from contextlib import suppress
from os.path import getmtime
from pathlib import Path
from time import time
from typing import cast, Any
from urllib.parse import quote

import aiofiles
import aiohttp
from geopy import Point

from betty.app import App
from betty.asyncio import gather
from betty.concurrent import RateLimiter
from betty.functools import filter_suppress
from betty.locale import Localized, negotiate_locale, to_locale, get_data, LocaleNotFoundError, Localey
from betty.media_type import MediaType
from betty.model.ancestry import Link, HasLinks, Place, File, HasFiles


class WikipediaError(BaseException):
    pass


class NotAPageError(WikipediaError, ValueError):
    pass


class RetrievalError(WikipediaError, RuntimeError):
    pass


_URL_PATTERN = re.compile(r'^https?://([a-z]+)\.wikipedia\.org/wiki/([^/?#]+).*$')


def _parse_url(url: str) -> tuple[str, str]:
    match = _URL_PATTERN.fullmatch(url)
    if match is None:
        raise NotAPageError
    return cast(tuple[str, str], match.groups())


class Summary(Localized):
    def __init__(self, locale: str, name: str, title: str, content: str):
        super().__init__(locale=locale)
        self._name = name
        self._title = title
        self._content = content

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return 'https://%s.wikipedia.org/wiki/%s' % (self.locale, self._name)

    @property
    def title(self) -> str:
        return self._title

    @property
    def content(self) -> str:
        return self._content


class Image:
    def __init__(
        self,
        path: Path,
        media_type: MediaType,
        title: str,
        wikimedia_commons_url: str,
    ):
        self._path = path
        self._media_type = media_type
        self._title = title
        self._wikimedia_commons_url = wikimedia_commons_url

    @property
    def path(self) -> Path:
        return self._path

    @property
    def media_type(self) -> MediaType:
        return self._media_type

    @property
    def title(self) -> str:
        return self._title

    @property
    def wikimedia_commons_url(self) -> str:
        return self._wikimedia_commons_url


class _Retriever:
    def __init__(
        self,
        http_client: aiohttp.ClientSession,
        cache_directory_path: Path,
        # Default to seven days.
        ttl: int = 86400 * 7,
    ):
        self._cache_directory_path = cache_directory_path
        self._cache_directory_path.mkdir(exist_ok=True, parents=True)
        self._ttl = ttl
        self._http_client = http_client
        self._rate_limiter = RateLimiter(200)
        self._images: dict[str, Image | None] = {}

    async def _request(self, url: str, extension: str | None = None) -> Any:
        cache_file_path = self._cache_directory_path / hashlib.md5(url.encode("utf-8")).hexdigest()
        if extension:
            cache_file_path = cache_file_path.with_suffix(f'.{extension}')

        response_data = None
        with suppress(FileNotFoundError):
            if getmtime(cache_file_path) + self._ttl > time():
                async with aiofiles.open(cache_file_path, mode='r+b') as f:
                    response_data = await f.read()

        if response_data is None:
            logger = logging.getLogger(__name__)
            try:
                logger.debug(f'Fetching {url}...')
                async with self._rate_limiter:
                    async with self._http_client.get(url) as response:
                        response_data = await response.read()
                async with aiofiles.open(cache_file_path, 'w+b') as f:
                    await f.write(response_data)
            except aiohttp.ClientError as error:
                logger.warning(f'Could not successfully connect to Wikipedia at {url}: {error}')
            except asyncio.TimeoutError:
                logger.warning(f'Timeout when connecting to Wikipedia at {url}')

        if response_data is None:
            try:
                async with aiofiles.open(cache_file_path, mode='r+b') as f:
                    response_data = await f.read()
            except FileNotFoundError:
                raise RetrievalError('Could neither fetch %s, nor find an old version in the cache.' % url)

        return response_data

    async def _get_query_api_data(self, url: str) -> dict[str, Any]:
        api_data = json.loads(await self._request(url))
        try:
            return api_data['query']['pages'][0]  # type: ignore[no-any-return]
        except (LookupError, TypeError) as error:
            raise RetrievalError(f'Could not successfully parse the JSON format returned by {url}: {error}')

    async def _get_page_query_api_data(self, page_language: str, page_name: str) -> dict[str, Any]:
        return await self._get_query_api_data(
            f'https://{page_language}.wikipedia.org/w/api.php?action=query&titles={quote(page_name)}&prop=langlinks|pageimages|coordinates&lllimit=500&piprop=name&pilicense=free&pilimit=1&coprimary=primary&format=json&formatversion=2'
        )

    async def get_translations(self, page_language: str, page_name: str) -> dict[str, str]:
        try:
            api_data = await self._get_page_query_api_data(page_language, page_name)
        except RetrievalError as error:
            logger = logging.getLogger(__name__)
            logger.warning(str(error))
            return {}
        try:
            translations_data = api_data['langlinks']
        except KeyError:
            # There may not be any translations.
            return {}
        return {translation_data['lang']: translation_data['title'] for translation_data in translations_data}

    async def get_summary(self, page_language: str, page_name: str) -> Summary | None:
        logger = logging.getLogger(__name__)
        try:
            url = f'https://{page_language}.wikipedia.org/api/rest_v1/page/summary/{page_name}'
            request_data = await self._request(url)
            try:
                api_data = json.loads(request_data)
                return Summary(
                    page_language,
                    page_name,
                    api_data['titles']['normalized'],
                    api_data['extract_html'] if 'extract_html' in api_data else api_data['extract'],
                )
            except (json.JSONDecodeError, KeyError) as error:
                raise RetrievalError(f'Could not successfully parse the JSON content returned by {url}: {error}')
        except RetrievalError as error:
            logger.warning(str(error))
        return None

    async def get_image(self, page_language: str, page_name: str) -> Image | None:
        try:
            api_data = await self._get_page_query_api_data(page_language, page_name)
            try:
                page_image_name = api_data['pageimage']
            except KeyError:
                # There may not be any images.
                return None

            if page_image_name in self._images:
                return self._images[page_image_name]

            url = f'https://en.wikipedia.org/w/api.php?action=query&prop=imageinfo&titles=File:{quote(page_image_name)}&iiprop=url|mime|canonicaltitle&format=json&formatversion=2'
            image_info_api_data = await self._get_query_api_data(url)

            try:
                image_info = image_info_api_data['imageinfo'][0]
            except KeyError as error:
                raise RetrievalError(f'Could not successfully parse the JSON content returned by {url}: {error}')

            extension = None
            for mimetypes_extension, mimetypes_media_type in mimetypes.types_map.items():
                if mimetypes_media_type == image_info['mime']:
                    extension = mimetypes_extension
            await self._request(image_info['url'], extension)

            file_path = (self._cache_directory_path / hashlib.md5(image_info['url'].encode("utf-8")).hexdigest()).with_suffix(f'.{extension}')
            image = Image(
                file_path,
                MediaType(image_info['mime']),
                # Strip "File:" or any translated equivalent from the beginning of the image's title.
                image_info['canonicaltitle'][image_info['canonicaltitle'].index(':') + 1:],
                image_info['descriptionurl'],
            )

            return image
        except RetrievalError as error:
            logger = logging.getLogger(__name__)
            logger.warning(str(error))
            return None

    async def get_place_coordinates(self, page_language: str, page_name: str) -> Point | None:
        api_data = await self._get_page_query_api_data(page_language, page_name)
        try:
            coordinates = api_data['coordinates'][0]
        except KeyError:
            # There may not be any coordinates.
            return None
        try:
            if coordinates['globe'] != 'earth':
                return None
            return Point(coordinates['lat'], coordinates['lon'])
        except KeyError as error:
            raise RetrievalError(f'Could not successfully parse the JSON content: {error}')


class _Populator:
    def __init__(self, app: App, retriever: _Retriever):
        self._app = app
        self._retriever = retriever
        self._image_files: dict[Image, File] = {}

    async def populate(self) -> None:
        locales = list(map(lambda x: x.alias, self._app.project.configuration.locales.values()))
        await gather(*(
            self._populate_entity(entity, locales)
            for entity
            in self._app.project.ancestry
            if isinstance(entity, HasLinks)
        ))

    async def _populate_entity(self, entity: HasLinks, locales: Sequence[str]) -> None:
        await self._populate_has_links(entity, locales)

        if isinstance(entity, HasFiles):
            await self._populate_has_files(entity)

        if isinstance(entity, Place):
            await self._populate_place(entity)

    async def _populate_has_links(self, has_links: HasLinks, locales: Sequence[str]) -> None:
        summary_links: MutableSequence[tuple[str, str]] = []
        for link in has_links.links:
            try:
                page_language, page_name = _parse_url(link.url)
            except NotAPageError:
                continue
            else:
                try:
                    get_data(page_language)
                except LocaleNotFoundError:
                    continue
                else:
                    summary_links.append((page_language, page_name))

            summary = None
            if link.label is None:
                with suppress(RetrievalError):
                    summary = await self._retriever.get_summary(page_language, page_name)
            await self.populate_link(link, page_language, summary)
        await self._populate_has_links_with_translation(has_links, locales, summary_links)

    async def _populate_has_links_with_translation(self, has_links: HasLinks, locales: Sequence[str], summary_links: MutableSequence[tuple[str, str]]) -> None:
        for page_language, page_name in summary_links:
            page_translations = await self._retriever.get_translations(page_language, page_name)
            if len(page_translations) == 0:
                continue
            page_translation_locale_datas: Sequence[Localey] = list(filter_suppress(get_data, LocaleNotFoundError, page_translations.keys()))
            for locale in locales:
                if locale == page_language:
                    continue
                added_page_locale_data = negotiate_locale(locale, page_translation_locale_datas)
                if added_page_locale_data is None:
                    continue
                added_page_language = to_locale(added_page_locale_data)
                added_page_name = page_translations[added_page_language]
                if (added_page_language, added_page_name) in summary_links:
                    continue
                added_summary = await self._retriever.get_summary(added_page_language, added_page_name)
                if not added_summary:
                    continue
                added_link = Link(added_summary.url)
                await self.populate_link(added_link, added_page_language, added_summary)
                has_links.links.append(added_link)
                summary_links.append((added_page_language, added_page_name))
            return

    async def populate_link(self, link: Link, summary_language: str, summary: Summary | None = None) -> None:
        if link.url.startswith('http:'):
            link.url = 'https:' + link.url[5:]
        if link.media_type is None:
            link.media_type = MediaType('text/html')
        if link.relationship is None:
            link.relationship = 'external'
        if link.locale is None:
            link.locale = summary_language
        if link.description is None:
            # There are valid reasons for links in locales that aren't supported.
            with suppress(ValueError):
                link.description = (await self._app.localizers.get_negotiated(link.locale))._('Read more on Wikipedia.')
        if summary is not None and link.label is None:
            link.label = summary.title

    async def _populate_place(self, place: Place) -> None:
        await self._populate_place_coordinates(place)

    async def _populate_place_coordinates(self, place: Place) -> None:
        if place.coordinates:
            return

        for link in place.links:
            try:
                page_language, page_name = _parse_url(link.url)
            except NotAPageError:
                continue
            else:
                coordinates = await self._retriever.get_place_coordinates(page_language, page_name)
                if coordinates:
                    place.coordinates = coordinates
                    return

    async def _populate_has_files(self, has_files: HasFiles & HasLinks) -> None:
        for link in has_files.links:
            try:
                page_language, page_name = _parse_url(link.url)
            except NotAPageError:
                continue
            else:
                image = await self._retriever.get_image(page_language, page_name)
                if not image:
                    continue

                try:
                    file = self._image_files[image]
                except KeyError:
                    file = File(
                        id=f'wikipedia-{image.title}',
                        path=image.path,
                        media_type=image.media_type,
                        links=[
                            Link(
                                f'{image.wikimedia_commons_url}?uselang={locale_configuration.alias}',
                                label=self._app.localizers[locale_configuration.locale]._('Description, licensing, and image history'),
                                description=self._app.localizers[locale_configuration.locale]._('Find out more about this image on Wikimedia Commons.'),
                                locale=locale_configuration.locale,
                                media_type=MediaType('text/html'),
                            )
                            for locale_configuration
                            in self._app.project.configuration.locales.values()
                        ],
                    )
                    self._image_files[image] = file

                has_files.files.add(file)
                self._app.project.ancestry.add(file)
                return



"""
Provide utilities for raising warnings.
"""
import warnings

import typing_extensions


class BettyDeprecationWarning(DeprecationWarning):
    """
    Raised for deprecated Betty functionality.
    """

    pass


def deprecate(message: str, stacklevel: int = 1):
    """
    Raise a Betty deprecation warning.
    """
    warnings.warn(message, category=BettyDeprecationWarning, stacklevel=stacklevel + 1)


class deprecated(typing_extensions.deprecated):
    """
    Decorate a class, function, or overload to indicate that it is depreated.

    This is identical to :py:class:`warnings.deprecated`, but raises a Betty
    deprecation warning.
    """

    def __init__(self, message: str, stacklevel: int = 1):
        super().__init__(message, category=BettyDeprecationWarning, stacklevel=stacklevel)
